from pyBittle.bittleManager import *
from pyBittle.bluetoothManager import *
from pyBittle.serialManager import *
from pyBittle.wifiManager import *

__author__ = "EnriqueMoran"

__version__ = "1.1.3"